/*
 *	Copyright 1990 by Rayan S. Zachariassen, all rights reserved.
 *	This will be free software, but only when it is finished.
 */

#include "mailer.h"

/* Start of the scanner */

/* scanner tables: definition of character classes according to RFC822 */

#define _h	01	/* header field */
#define _w	02	/* linear white-space character (space / htab) */
#define _d	04	/* digit */
#define _c	010	/* control */
#define _a	020	/* alphabetic */
#define _l	040	/* line feed */
#define _r	0100	/* carriage return */
#define _s	0200	/* specials */

/* ISO Latin 1 (8859) */

char	ctype[256] = {						/* octalcode */
_c,	_c,	_c,	_c,	_c,	_c,	_c,	_c,	/*   0 -   7 */
_c,	_c|_w,	_l|_c,	_c,	_c,	_r|_c,	_c,	_c,	/*  10 -  17 */
_c,	_c,	_c,	_c,	_c,	_c,	_c,	_c,	/*  20 -  27 */
_c,	_c,	_c,	_c,	_c,	_c,	_c,	_c,	/*  30 -  37 */
_w,	_h,	_s|_h,	_h,	_h,	_h,	_h,	_h,	/*  40 -  47 */
_s|_h,	_s|_h,	_h,	_h,	_s|_h,	_h,	_s|_h,	_h,	/*  50 -  57 */
_d|_h,	_d|_h,	_d|_h,	_d|_h,	_d|_h,	_d|_h,	_d|_h,	_d|_h,	/*  60 -  67 */
_d|_h,	_d|_h,	_s,	_s|_h,	_s|_h,	_h,	_s|_h,	_h,	/*  70 -  77 */
_s|_h,	_a|_h,	_a|_h,	_a|_h,	_a|_h,	_a|_h,	_a|_h,	_a|_h,	/* 100 - 107 */
_a|_h,	_a|_h,	_a|_h,	_a|_h,	_a|_h,	_a|_h,	_a|_h,	_a|_h,	/* 110 - 117 */
_a|_h,	_a|_h,	_a|_h,	_a|_h,	_a|_h,	_a|_h,	_a|_h,	_a|_h,	/* 120 - 127 */
_a|_h,	_a|_h,	_a|_h,	_s|_h,	_s|_h,	_s|_h,	_h,	_h,	/* 130 - 137 */
_h,	_a|_h,	_a|_h,	_a|_h,	_a|_h,	_a|_h,	_a|_h,	_a|_h,	/* 140 - 147 */
_a|_h,	_a|_h,	_a|_h,	_a|_h,	_a|_h,	_a|_h,	_a|_h,	_a|_h,	/* 150 - 157 */
_a|_h,	_a|_h,	_a|_h,	_a|_h,	_a|_h,	_a|_h,	_a|_h,	_a|_h,	/* 160 - 167 */
_a|_h,	_a|_h,	_a|_h,	_h,	_h,	_h,	_h,	_c,	/* 170 - 177 */
	/* The class assignments in the second half are guesses */
_c,	_c,	_c,	_c,	_c,	_c,	_c,	_c,	/* 200 - 207 */
_c,	_c,	_c,	_c,	_c,	_c,	_c,	_c,	/* 210 - 217 */
_c,	_c,	_c,	_c,	_c,	_c,	_c,	_c,	/* 220 - 227 */
_c,	_c,	_c,	_c,	_c,	_c,	_c,	_c,	/* 230 - 237 */
_a,	_a,	_a,	_a,	_a,	_a,	_a,	_a,	/* 240 - 247 */
_a,	_a|_h,	_a,	_s|_h,	_a,	_a|_h,	_a,	_a,	/* 250 - 257 */
_a,	_a,	_a,	_a,	_a,	_a,	_a,	_a,	/* 260 - 267 */
_a,	_a,	_a,	_s|_h,	_a,	_a,	_a,	_a,	/* 270 - 277 */
_a|_h,	_a|_h,	_a|_h,	_a|_h,	_a|_h,	_a|_h,	_a|_h,	_a|_h,	/* 300 - 307 */
_a|_h,	_a|_h,	_a|_h,	_a|_h,	_a|_h,	_a|_h,	_a|_h,	_a|_h,	/* 310 - 317 */
_a|_h,	_a|_h,	_a|_h,	_a|_h,	_a|_h,	_a|_h,	_a|_h,	_a|_h,	/* 320 - 327 */
_a|_h,	_a|_h,	_a|_h,	_a|_h,	_a|_h,	_a|_h,	_a|_h,	_a|_h,	/* 330 - 337 */
_a|_h,	_a|_h,	_a|_h,	_a|_h,	_a|_h,	_a|_h,	_a|_h,	_a|_h,	/* 340 - 347 */
_a|_h,	_a|_h,	_a|_h,	_a|_h,	_a|_h,	_a|_h,	_a|_h,	_a|_h,	/* 350 - 357 */
_a|_h,	_a|_h,	_a|_h,	_a|_h,	_a|_h,	_a|_h,	_a|_h,	_a|_h,	/* 360 - 367 */
_a|_h,	_a|_h,	_a|_h,	_a|_h,	_a|_h,	_a|_h,	_a|_h,	_a|_h	/* 370 - 377 */
};


/*
 * Tell whether we are looking at a new header line, a continuation line,
 * or if we are done with the header. Return the number of characters in
 * the name of the header, or 0 if a continuation, or < 0 if end of header.
 * If non-0, the cardinality of the number returned is the length of the
 * header field name.
 */

int
hdr_status(cp, lbuf, n)
	register u_char	*cp, *lbuf;
	int	n;
{
	if (*cp == ' ' || *cp == '\t') {
		while ((cp < lbuf + n) && (ctype[*cp&0177] & (_w|_l)))
			++cp;
		if (cp == lbuf + n)
			/* a line containing only whitespace is EOH */
			return -1;
		/* a continuation line (folded header) */
		return 0;
	}
	while ((cp < lbuf + n) && (ctype[*cp&0177] & _h))
		++cp;
	if (cp < lbuf + n && *cp == ':')	/* header line */
		return cp - lbuf;
	/* if we get to here, we have a malformed header line */
	return lbuf - cp;
}

#define MKERROR(msg,prev)	tn = makeToken((u_char *)(msg), strlen(msg)); \
				tn->t_type = Error; \
				tn->t_next = (prev); \
				(prev) = tn;

/*
 * Recognize a compound token, or rather, a token which is defined by
 * matching start and end delimiters. A comment or quoted string is
 * the typical example. Comments may be recursive.
 */

u_long
_hdr_compound(cp, n, cstart, cend, type, tp, tlist, tlistp)
	register u_char	*cp;
	long	n;
	u_char	cstart, cend;
	TokenType	type;
	struct token	*tp, **tlist, **tlistp;
{
	int i;
	struct token *tn;
	char msgbuf[50];

	if (*cp != cstart)
		abort();
nextline:
	while (n > 0 && --n > 0 && *++cp != cend) {
		if (*cp == cstart && type == Comment) {
			i = n;
			n = _hdr_compound(cp, n, cstart, cend,
					     type, tp, tlist, tlistp);
			if (n > 0) ++n;
			cp += (i - n);
		} else if (*cp == '\\') {
			if (--n == 0) {
				MKERROR("missing character after backslash",
						 *tlist);
				break;
			}
			++cp;
		} else if (*cp == '\r') {
			/* type = Error; */
			MKERROR("illegal CR in token",*tlist);
		}
	}
	if ((n == 0 && *(cp+1) != cend) || (n != 0 && *cp != cend)) {
		if (n == 0 && tlistp != NULL
		    && *tlistp != NULL && (*tlistp)->t_next != NULL) {
			/* compound token is continued on next line */
			*tlistp = (*tlistp)->t_next;
			n = TOKENLEN(*tlistp);
			cp = (*tlistp)->t_pname - 1 /* ack! pft! */;
			goto nextline;
		}
		/* type = Error; */	/* hey, no reason to refuse a message */
		(void) sprintf(msgbuf, "missing closing '%c' in token", cend);
		MKERROR(msgbuf,*tlist);
		tp->t_pname = 0;	/* ugly way of signalling scanner */
	} else
		--n;
	tp->t_type = type;
	return n;
}

/* Unfold (see RFC822) the contents of a compound token */

u_char *
_unfold(start, end, t)
	u_char *start, *end;
	struct token *t;
{
	u_char *s, *cp;
	/*
	 * Since we will be ignoring CRLF and copying everything else,
	 * we know the max length is going to be (end - start) (sic).
	 * However, I like being conservative about these things...
	 */

	s = cp = (u_char *)tmalloc((u_int)(end-start+1));
	while (start < end) {		/* < because *end is ')' or '"' */
		if (*start == '\n') {
			++start;
			continue;
		} else if (*start == '\0') {	/* lines are NUL terminated */
			t = t->t_next;		/* cannot be NULL! */
			start = t->t_pname;
		}
		*s++ = *start++;
	}
	*s = '\0';
	return cp;
}

/*
 * The Scanner.
 *
 * cpp		- pointer to pointer to string.
 * n		- number of characters left in string.
 * c1, c2	- if non-NUL, these characters should be considered Special.
 *
 * The scanner will return a token list corresponding to the n next characters
 * in the string. Originally only a single token was returned per call, but
 * for efficiency this was changed to avoid function call overhead. The tokens
 * returned are classified by type (TokenType enum class).
 */

struct token *
scan822(cpp, n, c1, c2, allowcomments, tlistp)
	u_char	**cpp;			/* pointer to pointer to text */
	long	n;			/* number of characters to scan */
	u_char	c1, c2;			/* temporary specials */
	int allowcomments;		/* #prefix tokens are comments to EOT */
	struct token **tlistp;		/* continuation line tokens if any */
{
	register u_char	*cp;
	static struct token	t;
	struct token	*tlist, *tp, *tn, *ot;
	char	ct, sc1, sc2, msgbuf[50];

	if (n == 0)
		return NULL;
	sc1 = sc2 = '\0';
	c1 &= 0177;
	if (c1 != '\0') {
		sc1 = ctype[c1];
		ctype[c1] |= _s;
	}
	c2 &= 0177;
	if (c2 != '\0') {
		sc2 = ctype[c2];
		ctype[c2] |= _s;
	}
	tlist = NULL;
	do {
		cp = *cpp;
		ct = ctype[*cp&0177];
		t.t_len = n;
		t.t_pname = cp;
		if (ct & _w) {		/* LWSP without the CR LF part */
			while (--n > 0 && (ctype[*++cp&0177] & _w))
				continue;
			t.t_type = Space;
		} else if (ct & _r) {	/* >= 1 CR followed by LFs is a fold */
			while (--n > 0 && (ctype[*++cp&0177] & _r))
				continue;
			if (n == 0 || !(ctype[*cp&0177] & _l)) {
				(void) strcpy(msgbuf, "missing newline (LF)");
				MKERROR(msgbuf,tlist);
			} else if (n > 1 && (ctype[*cp&0177] & _l)) {
				while (--n > 0 && (ctype[*++cp&0177] & _l))
					continue;
				(void) strcpy(msgbuf,
					"too many newlines (LFs) in field");
				MKERROR(msgbuf,tlist);
			}
			t.t_type = Fold;
		} else if (ct & _l) {	/* >= 1 LFs without CR is a fold too */
			while (--n > 0 && (ctype[*++cp&0177] & _l))
				continue;
			(void)strcpy(msgbuf,"too many newlines (LFs) in field");
			MKERROR(msgbuf,tlist);
			t.t_type = Fold;
		} else if ((ct & _s) && (*cp=='\050' || *cp=='"' || *cp=='[')) {
			TokenType	type;
			u_char	cend;

			if (*cp == '"') {
				cend = *cp;
				type = String;
			} else if (*cp == '[') {
				cend = ']';
				type = DomainLiteral;
			} else {
				cend = ')';
				type = Comment;
			}
			ot = (tlistp == NULL ? NULL : *tlistp);
			n = _hdr_compound(cp, n, *cp, cend,
					     type, &t, &tlist, tlistp);
			if (ot != NULL && tlistp != NULL && ot != *tlistp) {
				/* a compound token crossed line boundary */
				(*cpp) = (*tlistp)->t_pname +
					TOKENLEN(*tlistp) - (n+1);
				/* copy from ++cp up to and including *cpp */
				t.t_pname = _unfold(++cp,
					 n >= 0 ? (*cpp)-1 : (*cpp), ot);
				/* +/- n so t.t_len comes out correct below */
				t.t_len = strlen((char *)t.t_pname) + n;
				(*cpp) -= (t.t_len - n);
			} else {
				if (t.t_pname != NULL)
					/* magic sign, no ending char */
					--t.t_len, ++(*cpp);
				/* past first bracketing char */
				--t.t_len, ++(*cpp);
				t.t_pname = ++cp;
			}
		} else if (ct & _s) {		/* specials */
			--n;
			t.t_type = Special;
		} else if (!(ct & _c)) {	/* atom */
			while (--n > 0 && !(ctype[*++cp&0177]&(_w|_s|_c|_l|_r)))
				continue;
			t.t_type = Atom;
		} else {
			while (--n > 0 && (ctype[*++cp&0177] & _c))
				continue;
			(void) strcpy(msgbuf, "illegal control character");
			if (t.t_len > n+1)
				(void) strcat(msgbuf, "s");
			MKERROR(msgbuf,tlist);
			t.t_type = Atom;
		}
		t.t_len -= n;
		/* return two values */
		*cpp += t.t_len;
		if (allowcomments && t.t_len >= 1 && t.t_pname[0] == '#') {
			*cpp += n;
			break;
		}
		t.t_next = tlist;
		if (t.t_len > 0)
			tlist = copyToken(&t);
		else {
			t.t_pname = (u_char *)"";
			t.t_len = 0;
			tlist = copyToken(&t);
		}
	} while (n > 0);

	tp = tn = NULL;
	for (tp = NULL; tlist != NULL; tlist = tn) {
		tn = tlist->t_next;
		tlist->t_next = tp;
		tp = tlist;
	}
	if (c1 != '\0') ctype[c1] = sc1;
	if (c2 != '\0') ctype[c2] = sc2;
	return tp;
}
